/* resolve.c  acb  10-9-1994
 * routines for recursively resolving a RTN.
 */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include "rtn.h"
#include "resolve.h"
#include "dstring.h"
#include "map.h"
#include "transform.h"

#ifndef __GNUC__
#define inline /* no inline functions available */
#endif

struct _resolve_iter_params {
  pRule rules;
  char **result;
  pParam formal; /* the rule's formal parameters, as defined */
  pParam cooked; /* cooked */
};

extern int trace;
extern pListNode transformations;

static char *_resolve_rule(pRule rules, pRule rule, pNode param, 
			   pListNode parent_formal, pListNode parent_cooked);

/* resolve an atom */
static inline char *resolve_atom(pNode atom, pRule rules, pListNode formal, pListNode cooked)
{
  int index;

  if(trace) printf("Resolving atom: type = %d, data = \"%s\"\n",
		   atom->type, atom->data);
  switch(atom->type) {
  case literal:
    return strdup(atom->data);
  case symbol:
    /* first, check if this symbol is a parameter of the current rule */
    if((index=param_indexof(formal, atom->data))!=-1) {
      if(index>=list_length(cooked))
	return(strdup("(parameter off end of list)"));
      else
	return(strdup(list_nth(cooked, index)->data));
    };
    /* it is not a parameter substitution; resolve from rules */
    return _resolve_rule(rules, rule_find(rules, atom->data), atom->params,
			 formal, cooked);
  case mapping:
      if(map_lookup(mappings, atom->data))
	  return apply_mapping(resolve_atom(atom->params, rules, formal, 
					    cooked),
			       map_lookup(mappings, atom->data));
      return apply_xform(resolve_atom(atom->params, rules, formal, cooked),
			 trans_lookup(transformations, atom->data));
  case deref:
    return _resolve_rule(rules, 
			 rule_find(rules, resolve_atom(atom->params, rules, 
						       formal, cooked)),
			 NULL, formal, cooked);
  case var_conddef: {
    char *val = var_fetch(atom->data);
    if(!val) {
      val = resolve_atom(atom->params, rules, formal, cooked);
      var_put(atom->data, val);
    };
    val = strdup(val); /* make a disposable copy of it */
    return val;
  };
  case var_def: {
      char *val = resolve_atom(atom->params, rules, formal, cooked);
      if(!val) val = "undefined";
      var_put(atom->data, val);
      val = strdup(val); /* make a disposable copy of it */
      return val;
  };
  case var_ref: {
      char *val = var_fetch(atom->data);      
      if(!val) {
	  fprintf(stderr, "undefined variable `%s'.\n", atom->data);
	  exit(1);
      };
      return strdup(val); /* make a disposable copy of it */      
  };
  case silence: {
    free(resolve_atom(atom->params, rules, formal, cooked));
    return NULL;
  };
  case code: return exec_stream(atom->data);
  };
};

static void resolve_iterator(pNode atom, struct _resolve_iter_params *prms)
{
  *(prms->result) = dstrcat(*(prms->result), 
			    resolve_atom(atom, prms->rules, 
					 prms->formal, prms->cooked));
};

static inline char *resolve_option(pRule rules, pOption option, 
  pParam formal, pParam cooked)
{
  struct _resolve_iter_params prms;
  char *result=NULL;
  prms.rules = rules;
  prms.result = &result;
  prms.formal = formal;
  prms.cooked = cooked;
  node_map(option->atoms, &resolve_iterator, &prms);
  return result;
};

/* resolve a parameter (node) list into a list of strings; now checks a
   list of parameters from the parent. */

static pListNode resolve_params(pNode in, pRule rules, 
				pListNode formal, pListNode cooked)
{
  int index;
  if(in==(pNode)NULL) return (pParam)NULL;
  /* check if it's a cascading parameter */
  if((index=param_indexof(formal, in->data))!=-1) {
    char *this_result;
    this_result = (index>=list_length(cooked))?"(parameter off end of list)"
      :list_nth(cooked, index);
    return list_cons(strdup(this_result), 
		     resolve_params(in->next, rules, formal, cooked));
  };
  return list_cons((void *)resolve_atom(in, rules,
					rule_find(rules, in->data)?
					rule_find(rules, in->data)->params:
					NULL,
					resolve_params(in->params, rules,
						       formal, cooked)),
		   resolve_params(in->next, rules, formal, cooked));
};

/* resolve the rule rule recursively, using rules as the rule-base.
   This function now accepts the parent's formal and actual rules. */

static char *_resolve_rule(pRule rules, pRule rule, pNode param,
			   pListNode parent_formal, pListNode parent_cooked)
{
  if(rule) {
    char *r;
    pParam cooked = resolve_params(param, rules, parent_formal, parent_cooked);
    int num_options = option_length(rule->options);
    int choice;

    if(trace) printf("Resolving rule <%s>\n", rule->symbol);
    /* make the choice */
    do {
      choice = random()%num_options;
    } while ((choice==rule->last_choice) && (num_options>1));
    rule->last_choice = choice;

    r = resolve_option(rules, option_nth(rule->options, choice),
		       rule->params, cooked);
    list_free(cooked, free_car_destructor);
    if(trace) printf("resolve_option returned \"%s\"\n", r);
    return r;
  } else return NULL;
};

char *resolve_rule(pRule rules, pRule rule)
{
  return _resolve_rule(rules, rule, NULL, NULL, NULL);
};




